/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2005 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

//mlsTrace.cpp
//Multithreaded load instruction tracing

// Bugs and criticisms can be directed to
// myers.albert@gmail.com

#include <unistd.h>
#include <stdio.h>
#include <iomanip>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <limits>
#include <algorithm>
#include <deque>
#include <sys/time.h>
#include <sys/stat.h>
#include "pin.H"
#include "instlib.H"
#include "mlsTraceAssist.h"
#include "mlsTrace.h"

//Can be used to filter which parts of the binary to instrument
INSTLIB::FILTER filter;

// Is called for every trace 
// Trace: Sequence of instructions that is always entered at the top and may have multiple exits.
VOID ASCIILoadTrace(TRACE trace, VOID *v)
{  
   //return if not a selected routine 
   if(!filter.SelectTrace(trace))
      return;

   //iterate over instructions
   for(BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl) )
   {
      for(INS ins = BBL_InsHead(bbl); INS_Valid(ins); ins = INS_Next(ins) )
      {
         INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)SetFastForwardAndLength, IARG_THREAD_ID, IARG_END);
  
         //If load instruction or store instruction ( can be both )
         if(INS_IsMemoryRead(ins) || INS_IsMemoryWrite(ins) )
         {  
            UINT32 memOperands = INS_MemoryOperandCount(ins);
            //iterate over operands
            for(UINT32 memOp = 0; memOp < memOperands; memOp++)
            {
               //if operand is load
               if( INS_MemoryOperandIsRead(ins, memOp) && TraceLoad.Value() )
               {
                  INS_InsertPredicatedCall(ins, IPOINT_BEFORE, (AFUNPTR)Emit_LoadValueDescriptor_ASCII, 
                                           IARG_THREAD_ID, IARG_INST_PTR, IARG_MEMORYOP_EA, memOp, IARG_MEMORYREAD_SIZE, IARG_END);
                                         //Arguments: TID, ins address, operand effec address , operand size in bytes                       
               }
               //if operand is store
               if(INS_MemoryOperandIsWritten(ins, memOp) && TraceStore.Value() )
               {
                  //protect store instruction with lock
                  INS_InsertPredicatedCall( ins, IPOINT_BEFORE, (AFUNPTR)lock_WriteLocation, IARG_FAST_ANALYSIS_CALL, IARG_THREAD_ID, IARG_MEMORYOP_EA, memOp, IARG_END);
                  
                  IPOINT where = IPOINT_AFTER;
                  if (!INS_HasFallThrough(ins))
                     where = IPOINT_TAKEN_BRANCH;
                     
                  INS_InsertPredicatedCall( ins, where, (AFUNPTR)Emit_StoreValueDescriptor_ASCII,
                                            IARG_THREAD_ID, IARG_INST_PTR, IARG_MEMORYWRITE_SIZE, IARG_END);
                                          //Arguments: TID, ins address, operand size in bytes                       
               }
            }
         }
      }
   }
}
VOID DisLoadTrace(TRACE trace, VOID *v)
{  
   //return if not a selected routine 
   if(!filter.SelectTrace(trace))
      return;

   //iterate over instructions
   for(BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl) )
   {
      for(INS ins = BBL_InsHead(bbl); INS_Valid(ins); ins = INS_Next(ins) )
      {
         INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)SetFastForwardAndLength, IARG_THREAD_ID, IARG_END);
  
         //If load instruction or store instruction ( can be both )
         if(INS_IsMemoryRead(ins) || INS_IsMemoryWrite(ins) )
         {  
            UINT32 memOperands = INS_MemoryOperandCount(ins);
            //iterate over operands
            for(UINT32 memOp = 0; memOp < memOperands; memOp++)
            {
               //if operand is load
               if( INS_MemoryOperandIsRead(ins, memOp) && TraceLoad.Value() )
               {
                  INS_InsertPredicatedCall(ins, IPOINT_BEFORE, (AFUNPTR)Emit_LoadValueDescriptor_Dis, 
                                           IARG_THREAD_ID, IARG_INST_PTR, IARG_MEMORYOP_EA, memOp, IARG_MEMORYREAD_SIZE, IARG_END);
                                         //Arguments: TID, ins address, operand effec address , operand size in bytes                       
               }
               //if operand is store
               if( INS_MemoryOperandIsWritten(ins, memOp) && TraceStore.Value() )
               {
                  //protect store instruction with lock
                  INS_InsertPredicatedCall( ins, IPOINT_BEFORE, (AFUNPTR)lock_WriteLocation, IARG_FAST_ANALYSIS_CALL, IARG_THREAD_ID, IARG_MEMORYOP_EA, memOp, IARG_END);
                  
                  IPOINT where = IPOINT_AFTER;
                  if (!INS_HasFallThrough(ins))
                     where = IPOINT_TAKEN_BRANCH;
                     
                  INS_InsertPredicatedCall( ins, where, (AFUNPTR)Emit_StoreValueDescriptor_Dis,
                                            IARG_THREAD_ID, IARG_INST_PTR, IARG_MEMORYWRITE_SIZE, IARG_END);
                                          //Arguments: TID, ins address, operand size in bytes                       

               }
            }
         }
      }
   }
}
VOID BinLoadTrace(TRACE trace, VOID *v)
{  
   //return if not a selected routine 
   if(!filter.SelectTrace(trace))
      return;

   //iterate over instructions
   for(BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl) )
   {
      for(INS ins = BBL_InsHead(bbl); INS_Valid(ins); ins = INS_Next(ins) )
      {
         INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)SetFastForwardAndLength, IARG_THREAD_ID, IARG_END);
  
         //If load instruction or store instruction ( can be both )
         if(INS_IsMemoryRead(ins) || INS_IsMemoryWrite(ins) )
         {  
            UINT32 memOperands = INS_MemoryOperandCount(ins);
            //iterate over operands
            for(UINT32 memOp = 0; memOp < memOperands; memOp++)
            {
               //if operand is load
               if( INS_MemoryOperandIsRead(ins, memOp) && TraceLoad.Value() )
               {
                  INS_InsertPredicatedCall(ins, IPOINT_BEFORE, (AFUNPTR)Emit_LoadValueDescriptor_Bin, 
                                           IARG_THREAD_ID, IARG_INST_PTR, IARG_MEMORYOP_EA, memOp, IARG_MEMORYREAD_SIZE, IARG_END);
                                         //Arguments: TID, ins address, operand effec address , operand size in bytes                       
               }
               //if operand is store
               if(INS_MemoryOperandIsWritten(ins, memOp) && TraceStore.Value() )
               {
                  //protect store instruction with lock
                  INS_InsertPredicatedCall( ins, IPOINT_BEFORE, (AFUNPTR)lock_WriteLocation, IARG_FAST_ANALYSIS_CALL, IARG_THREAD_ID, IARG_MEMORYOP_EA, memOp, IARG_END);
                  
                  IPOINT where = IPOINT_AFTER;
                  if (!INS_HasFallThrough(ins))
                     where = IPOINT_TAKEN_BRANCH;
                     
                  INS_InsertPredicatedCall( ins, where, (AFUNPTR)Emit_StoreValueDescriptor_Bin,
                                            IARG_THREAD_ID, IARG_INST_PTR, IARG_MEMORYWRITE_SIZE, IARG_END);
                                          //Arguments: TID, ins address, operand size in bytes                       
               }
            }
         }
      }
   }
}

VOID Fini(INT32 code, VOID *v)
{
   gettimeofday(&t2, NULL);
   //write remaining descriptors to file
   if(ASCII.Value())
   {
      if(usingCompression)
         WriteASCIIDescriptorTableToCompressor(PIN_GetTid());
      else
         WriteASCIIDescriptorTableToFile(PIN_GetTid());
   }
   else
   {
      if(usingCompression)
         WriteBinaryDescriptorTableToCompressor(PIN_GetTid());
      else
         WriteBinaryDescriptorTableToFile(PIN_GetTid());
   }
   //close 
   closeOutput();
   PrintStatistics();
}

//Called if Pin detaches from application
VOID DetachCallback(VOID *args)
{
   gettimeofday(&t2, NULL);
   //write remaining descriptors to file
   if(ASCII.Value())
   {
      if(usingCompression)
         WriteASCIIDescriptorTableToCompressor(PIN_GetTid());
      else
         WriteASCIIDescriptorTableToFile(PIN_GetTid());
   }
   else
   {
      if(usingCompression)
         WriteBinaryDescriptorTableToCompressor(PIN_GetTid());
      else
         WriteBinaryDescriptorTableToFile(PIN_GetTid());
   }

   //close file/pipe
   closeOutput();
   std::cout << "mlsTrace: Detaching..." << std::endl;
   PrintStatistics();
}

//called on thread creation
VOID ThreadStart(THREADID threadid, CONTEXT *ctxt, INT32 flags, VOID *v)
{
   PIN_GetLock(&cout_lock, threadid+1);
   UINT8* tid = new UINT8;
   *tid = numThreads;
   //This PIN tool spawns its own thread to write to file, which interrupts the logical ordering of threads
   //Use TLS to record application thread ids
   PIN_SetThreadData(tls_key, tid, threadid);
   std::cout << "mlsTrace: thread begin " << (UINT16) numThreads << " " << PIN_GetTid() << std::endl;
   numThreads++;
   PIN_ReleaseLock(&cout_lock);
}
//delete thread local storage
VOID ThreadFini(THREADID threadid, const CONTEXT *ctxt, INT32 flags, VOID *v)
{
   delete static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
}
//print knob usage on error
INT32 Usage()
{
   std::cerr << KnobOutputFile.StringKnobSummary() << std::endl;
   return -1;
}

int main(int argc, char * argv[])
{
   // Initialize pin
   PIN_InitSymbols();
   if (PIN_Init(argc, argv)) return Usage();

   if(KnobOutputFile.Value() == "mlsTrace.out")
   {
      //Create time stamp for output file so user won't overwrite previous files
      time_t t = time(NULL);
      stamp = *localtime(&t);
      string temp;
      temp = std::to_string(static_cast<long long>(stamp.tm_year + 1900)) + "_" + std::to_string(static_cast<long long>(stamp.tm_mon + 1))+ "_"
             + std::to_string(static_cast<long long>(stamp.tm_mday)) + "_" + std::to_string(static_cast<long long>(stamp.tm_hour)) + "."
             + std::to_string(static_cast<long long>(stamp.tm_min)) + "." + std::to_string(static_cast<long long>(stamp.tm_sec));
      //Concatenate file name and time stamp

      fileName = KnobOutputFile.Value() + temp;
   }
   else
      fileName = KnobOutputFile.Value();

   //Initialize locks
   PIN_InitLock(&table_lock);
   PIN_InitLock(&decoder_lock);
   PIN_InitLock(&cout_lock);
   PIN_InitLock(&count_lock);
   PIN_InitLock(&mem_lock);
   
   //The number of instructions to trace 
   if(Length.Value()==0)
      noLength = TRUE;
   else
      noLength = FALSE;
   
   //If writing descripts to binary file
   if( ASCII.Value()== FALSE )
   {
      if(AnnotateDisassembly.Value() == TRUE)
      {
         std::cout << "mlsTrace: Can only annotate disassembly if writing to ASCII file" << std::endl;
         std::exit(-1);
      }
      std::cout << "mlsTrace: Writing to binary file: " << fileName + ".bin" << std::endl;

      std::cout << "mlsTrace descriptor: ";
      std::cout << "ThreadID ( " << sizeof(UINT8) << " byte ), ";
      
      if( TraceLoad.Value() && TraceStore.Value() )
         std::cout << "Load/Store ( " << sizeof(UINT8) << " byte ), ";

      
      std::cout << "Instruction Address ( "  << sizeof(ADDRINT) << " bytes ), " 
	        << "Operand Address ( " << sizeof(ADDRINT) << " bytes ), " 
                << "Operand Size ( " << sizeof(UINT8) << " byte ), " 
	        << "Value ( Operand Size ) " << std::endl;
   }
   //Writing to Ascii
   else
   {
      std::cout << "mlsTrace: Writing to text file: " << fileName + ".txt" << std::endl;
      std::cout << "mlsTrace descriptor: ThreadID, ";
      if( TraceLoad.Value() && TraceStore.Value() )
         std::cout << "Load/Store, ";
      std::cout << " Instruction Address, Operand Address, Operand Size, Value"<< std::endl;

      //initialize decoder used to get instruction string
      if( AnnotateDisassembly.Value() == TRUE )
         xed_tables_init();
   }
   
   //Open Output file/pipe
   openOutput();

   //megabytes to bytes, go with power of ten for disk
   //will detach if larger than 
   fileLimit = FileLimit.Value() * 1000000;
   //Register Fini to be called when the application exits
   PIN_AddFiniFunction(Fini, NULL);
   //Register a detach function to be called if the pin exits early
   PIN_AddDetachFunction(DetachCallback, NULL);

   //ThreadStart is called when an application thread is created
   PIN_AddThreadStartFunction(ThreadStart, NULL);
   //used to delete local storage
   PIN_AddThreadFiniFunction(ThreadFini, NULL);

   //insert instrumentation procedures
   if( ASCII.Value() )
   {
      if( AnnotateDisassembly.Value() )
         TRACE_AddInstrumentFunction(DisLoadTrace, NULL);
      else
         TRACE_AddInstrumentFunction(ASCIILoadTrace, NULL);
   }
   else
      TRACE_AddInstrumentFunction(BinLoadTrace, NULL);

   // Activate procedure filter
   filter.Activate();
   // Start the program, never returns
   gettimeofday(&t1, NULL);
   PIN_StartProgram();
    
   return 0;
}
