/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2005 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

//Bugs and criticisms can be directed to
//myers.albert@gmail.com

//mlsTrace.h
//contains instruction analysis functions injected with mlsTrace.cpp

//procedures to reverse endianess
static inline VOID reverseOther(UINT8* valBuf, const UINT16 size)
{
    UINT8* end = valBuf + size;
    std::reverse(valBuf, end);
}
static inline VOID reverseHalfWord(UINT8* valBuf)
{
    UINT8 cTemp;
    cTemp= valBuf[1]; valBuf[1] = valBuf[0]; valBuf[0] = cTemp;
}
static inline VOID reverseWord(UINT8* valBuf)
{
    UINT8 cTemp;
    cTemp = valBuf[3]; valBuf[3] = valBuf[0]; valBuf[0] = cTemp;
    cTemp = valBuf[2]; valBuf[2] = valBuf[1]; valBuf[1] = cTemp;
}
static inline VOID reverseDoubleWord(UINT8* valBuf)
{
    UINT8 cTemp;
    cTemp = valBuf[7]; valBuf[7] = valBuf[0]; valBuf[0] = cTemp;
    cTemp = valBuf[6]; valBuf[6] = valBuf[1]; valBuf[1] = cTemp;
    cTemp = valBuf[5]; valBuf[5] = valBuf[2]; valBuf[2] = cTemp;
    cTemp = valBuf[4]; valBuf[4] = valBuf[3]; valBuf[3] = cTemp;
}
static inline VOID reverseExtendedPrecision(UINT8* valBuf)
{
    UINT8 cTemp;
    cTemp = valBuf[9]; valBuf[9] = valBuf[0]; valBuf[0] = cTemp; 
    cTemp = valBuf[8]; valBuf[8] = valBuf[1]; valBuf[1] = cTemp;
    cTemp = valBuf[7]; valBuf[7] = valBuf[2]; valBuf[2] = cTemp;
    cTemp = valBuf[6]; valBuf[6] = valBuf[3]; valBuf[3] = cTemp;
    cTemp = valBuf[5]; valBuf[5] = valBuf[4]; valBuf[4] = cTemp;
}
static inline VOID ConvertToBigEndian(UINT8* valBuf, UINT16 size)
{
    switch(size)
    {  
       //byte 
       case 1: 
          break;
       //half-word
       case 2:
          reverseHalfWord(valBuf);
          break;
       //word
       case 4:
          reverseWord(valBuf);
          break;
       //double word
       case 8:
          reverseDoubleWord(valBuf);
          break;
       //extended precision
       case 10:
          reverseExtendedPrecision(valBuf);
          break;
       default:
          break;
   }
}

//analysis calls to lock write instructions
//used to hold operand address for locked store operand
const ADDRINT *lockedOperandAddress;
VOID PIN_FAST_ANALYSIS_CALL lock_WriteLocation( const THREADID threadid, const ADDRINT * ea)
{
   PIN_GetLock(&mem_lock, threadid+1);
   lockedOperandAddress =  ea;
}

//Record ASCII load descriptor
VOID Emit_LoadValueDescriptor_ASCII(const THREADID threadid, const ADDRINT address, const ADDRINT * ea, const  UINT32 opSize)
{ 
   //if we can't record yet, return
   if( !CanEmit(threadid) ) return;
   //increment load statistics
   IncrementLoadStatistics(opSize);

   //copy value 
   UINT8 valBuf[opSize];
   PIN_SafeCopy(valBuf, ea, opSize);
   ConvertToBigEndian(valBuf, opSize);

   std::ostringstream convert;
   UINT8 assignedTID = *static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
   convert << std::dec << static_cast<UINT32>(assignedTID);
   if( TraceStore.Value() )
      convert << ", L";
   convert << ", 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << address;
   convert << ", " << std::internal << std::hex << std::setw(ADDRWID+2) << std::setfill('0') << ea;
   convert << ", " << std::dec << opSize;
   convert << ", 0x";
   
   for(UINT32 i = 0; i < opSize; i++)
   { 
      convert << std::hex << std::setw(2) << std::setfill('0') << static_cast<UINT32>(valBuf[i]);
   }
   convert << std::endl;

   string ASCIIDescriptor = convert.str();

   PIN_GetLock(&table_lock, threadid+1);
   //Push back, will write on close
   ASCIIDescriptorTable.push_back(ASCIIDescriptor);
   IncrementFileCount(ASCIIDescriptor.size());
   PIN_ReleaseLock(&table_lock);
}
VOID Emit_StoreValueDescriptor_ASCII(const THREADID threadid, const ADDRINT address, const  UINT32 opSize)
{
   //copy value 
   UINT8 valBuf[opSize];
   PIN_SafeCopy(valBuf, lockedOperandAddress, opSize);
   //release lock that was set before executing store instruction
   PIN_ReleaseLock(&mem_lock);
 
   
   //if we can't record yet, return
   if( !CanEmit(threadid) ) return;
   //increment load statistics
   IncrementStoreStatistics(opSize);

   ConvertToBigEndian(valBuf, opSize);

   std::ostringstream convert;
   UINT8 assignedTID = *static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
   convert << std::dec << static_cast<UINT32>(assignedTID);
   if( TraceLoad.Value() )
      convert << ", S";
   convert << ", 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << address;
   convert << ", " << std::internal << std::hex << std::setw(ADDRWID+2) << std::setfill('0') << lockedOperandAddress;
   convert << ", " << std::dec << opSize;
   convert << ", 0x";
   
   for(UINT32 i = 0; i < opSize; i++)
   { 
      convert << std::hex << std::setw(2) << std::setfill('0') << static_cast<UINT32>(valBuf[i]);
   }
   convert << std::endl;

   string ASCIIDescriptor = convert.str();

   PIN_GetLock(&table_lock, threadid+1);
   //Push back, will write on close
   ASCIIDescriptorTable.push_back(ASCIIDescriptor);
   IncrementFileCount(ASCIIDescriptor.size());
   PIN_ReleaseLock(&table_lock);
}
//Record ASCII descriptor with annotated disassembly
VOID Emit_LoadValueDescriptor_Dis(const THREADID threadid, const ADDRINT address, const ADDRINT * ea, const  UINT32 opSize)
{ 
   //if we can't record yet, return
   if( !CanEmit(threadid) ) return;
   //increment load statistics
   IncrementLoadStatistics(opSize);

   //get instruction string
   std::string ins;
   getAssemblyString(ins, threadid, address);
 
   //copy value 
   UINT8 valBuf[opSize];
   PIN_SafeCopy(valBuf, ea, opSize);
   ConvertToBigEndian(valBuf, opSize);

   std::ostringstream convert;
   UINT8 assignedTID = *static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
   convert << std::dec << static_cast<UINT32>(assignedTID);
   if( TraceStore.Value() )
      convert << ", L";
   convert << ", 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << address;
   convert << ", " << std::internal << std::hex << std::setw(ADDRWID+2) << std::setfill('0') << ea;
   convert << ", " << std::dec << opSize;
   convert << ", 0x";
   
   for(UINT32 i = 0; i < opSize; i++)
   { 
      convert << std::hex << std::setw(2) << std::setfill('0') << static_cast<UINT32>(valBuf[i]);
   }
   convert << "\t" + ins << std::endl;

   string ASCIIDescriptor = convert.str();

   PIN_GetLock(&table_lock, threadid+1);
   //Push back, will write on close
   ASCIIDescriptorTable.push_back(ASCIIDescriptor);
   IncrementFileCount(ASCIIDescriptor.size());
   PIN_ReleaseLock(&table_lock);
}
VOID Emit_StoreValueDescriptor_Dis(const THREADID threadid, const ADDRINT address, const  UINT32 opSize)
{
   //copy value 
   UINT8 valBuf[opSize];
   PIN_SafeCopy(valBuf, lockedOperandAddress, opSize);
   //release lock that was set before executing store instruction
   PIN_ReleaseLock(&mem_lock);
 
   
   //if we can't record yet, return
   if( !CanEmit(threadid) ) return;
   //increment load statistics
   IncrementStoreStatistics(opSize);

   //get instruction string
   std::string ins;
   getAssemblyString(ins, threadid, address);
 
   ConvertToBigEndian(valBuf, opSize);

   std::ostringstream convert;
   UINT8 assignedTID = *static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
   convert << std::dec << static_cast<UINT32>(assignedTID);
   if( TraceLoad.Value() )
      convert << ", S";
   convert << ", 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << address;
   convert << ", " << std::internal << std::hex << std::setw(ADDRWID+2) << std::setfill('0') << lockedOperandAddress;
   convert << ", " << std::dec << opSize;
   convert << ", 0x";
   
   for(UINT32 i = 0; i < opSize; i++)
   { 
      convert << std::hex << std::setw(2) << std::setfill('0') << static_cast<UINT32>(valBuf[i]);
   }
   convert << "\t" + ins << std::endl;

   string ASCIIDescriptor = convert.str();

   PIN_GetLock(&table_lock, threadid+1);
   //Push back, will write on close
   ASCIIDescriptorTable.push_back(ASCIIDescriptor);
   IncrementFileCount(ASCIIDescriptor.size());
   PIN_ReleaseLock(&table_lock);
}
//Record binary descriptor
VOID Emit_LoadValueDescriptor_Bin( const THREADID threadid, const ADDRINT address, const ADDRINT * ea, const UINT32 opSize)
{
   //if we can't record yet, return
   if( !CanEmit(threadid) ) return;
   //increment load statistics
   IncrementLoadStatistics(opSize);
 
   BinaryDescriptorTableEntry BinDescriptor;
   BinDescriptor.type = load;
   BinDescriptor.tid = *static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
   BinDescriptor.insAddr = address;

   BinDescriptor.operandEffAddr = reinterpret_cast<intptr_t>(ea); 
   BinDescriptor.operandSize = opSize;

   //copy value 
   UINT8 valBuf[opSize];
   PIN_SafeCopy(valBuf, ea, opSize);
   //reverse endianess
   //ConvertToBigEndian(valBuf, opSize);
 
  //Allocate memory for value
   BinDescriptor.data = new UINT8[opSize];
   //copy to struct entry
   std::copy(valBuf, valBuf+opSize, BinDescriptor.data); 
 
   //critical section
   PIN_GetLock(&table_lock, threadid+1);
   //Push back, will write on close
   binDescriptorTable.push_back(BinDescriptor);
   IncrementFileCount(BinaryDescriptorSize+opSize);
   PIN_ReleaseLock(&table_lock);
}

VOID Emit_StoreValueDescriptor_Bin( const THREADID threadid, const ADDRINT address, const UINT32 opSize)
{
   //copy value 
   UINT8 valBuf[opSize];
   PIN_SafeCopy(valBuf, lockedOperandAddress, opSize);
   PIN_ReleaseLock(&mem_lock);

   //if we can't record yet, return
   if( !CanEmit(threadid) ) return;
   //increment load statistics
   IncrementStoreStatistics(opSize);
 
   BinaryDescriptorTableEntry BinDescriptor;
   BinDescriptor.type = store;
   BinDescriptor.tid = *static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
   BinDescriptor.insAddr = address;
   BinDescriptor.operandEffAddr = reinterpret_cast<intptr_t>(lockedOperandAddress);
   BinDescriptor.operandSize = opSize;

   //reverse endianess
   //ConvertToBigEndian(valBuf, opSize);
     
   //Allocate memory for value
   BinDescriptor.data = new UINT8[opSize];
   //copy to struct entry
   std::copy(valBuf, valBuf+opSize, BinDescriptor.data); 
 
   //critical section
   PIN_GetLock(&table_lock, threadid+1);
   //Push back, will write on close
   binDescriptorTable.push_back(BinDescriptor);
   IncrementFileCount(BinaryDescriptorSize+opSize);
   PIN_ReleaseLock(&table_lock);
}
//used to decrement fastfoward_count and trace length for all instructions
VOID SetFastForwardAndLength(const THREADID threadid)
{
   //critical section
   PIN_GetLock(&count_lock, threadid+1);
   //if file limit is reacher or trace limit is reached
   //exit tool
   if( (traceLength + 1) == Length.Value() && !noLength )
   {
      PIN_Detach();
      PIN_ReleaseLock(&count_lock);
      return;
   }
   //decrement fast forward count   
   if( fastforwardLength < Skip.Value())
      fastforwardLength++;

   //otherwise increment length
   else
      traceLength++;

   PIN_ReleaseLock(&count_lock);
   return;
}


